﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion


using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TestFunctions.MultiObjective;
namespace HeuristicLab.Algorithms.MOCMAEvolutionStrategy {
  [Item("HypervolumeIndicator", "Selection of Offspring based on contributing Hypervolume")]
  [StorableClass]
  internal class HypervolumeIndicator : Item, IIndicator {

    public int LeastContributer<TR>(IEnumerable<TR> front, Func<TR, double[]> extractor, MultiObjectiveBasicProblem<RealVectorEncoding> problem) {
      var frontArr = front.Select(extractor.Invoke).ToList();
      if (frontArr.Count <= 1) return 0;
      var p = problem as MultiObjectiveTestFunctionProblem;
      var refPoint = BuildReference(p != null ? frontArr.Concat(new[] { p.TestFunction.ReferencePoint(p.Objectives) }) : frontArr, problem.Maximization);
      var hv = Hypervolume.Calculate(frontArr, refPoint, problem.Maximization);
      var contributions = Enumerable.Range(0, frontArr.Count).Select(i => Contribution(frontArr, i, problem.Maximization, refPoint, hv));
      return contributions.ArgMin();
    }

    private static double Contribution(IList<double[]> front, int idx, bool[] maximization, double[] refPoint, double hv) {
      var point = front[idx];
      front.RemoveAt(idx);
      var contribution = hv - Hypervolume.Calculate(front, refPoint, maximization);
      front.Insert(idx, point);
      return contribution;
    }
    private static double[] BuildReference(IEnumerable<double[]> front, IReadOnlyList<bool> maximization) {
      var refPoint = new double[maximization.Count];
      foreach (var point in front)
        for (var i = 0; i < maximization.Count; i++)
          refPoint[i] = maximization[i] ? Math.Min(refPoint[i], point[i]) : Math.Max(refPoint[i], point[i]);
      return refPoint;
    }

    [StorableConstructor]
    protected HypervolumeIndicator(bool deserializing) : base(deserializing) { }
    protected HypervolumeIndicator(HypervolumeIndicator original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) { return new HypervolumeIndicator(this, cloner); }
    public HypervolumeIndicator() { }
  }
}
